from __future__ import division

import numpy as np
from .cykmeans import cy_hikmeans, cy_hikmeans_push, algorithm_type_ikmeans
import math


def hikmeans(data, num_centers, num_leaves, algorithm='LLOYD', verbose=False):
    """
    Hierachical integer K-means

    Parameters
    ----------
    data : [N, D] `uint8` `ndarray`
        Data to be clustered
    num_centers : `int`
        Number of clusters (leaves) per level
    num_leaves : `int`
        Minimum number of final leaves in the tree. The depth of the tree
        is computed from this parameter and ``num_centers`
    algorithm : {'LLOYD', 'ELKAN'}, optional
        Algorithm to be used for clustering.
    verbose : `bool`, optional
        If ``True``, be verbose.

    Returns
    -------
    (tree_structure, assignments) : (object, [N, depth] `uint32` `ndarray`)
        Computed organisation and per-level assignments for ``data``
    """
    assert isinstance(data, np.ndarray)
    assert isinstance(num_centers, int)
    assert isinstance(num_leaves, int)
    assert isinstance(verbose, bool)
    if data.ndim != 2:
        raise ValueError('Data should be a 2-D matrix')

    if data.dtype != np.uint8:
        raise ValueError('Data should be uint8')

    if num_centers > data.shape[0]:
        raise ValueError('num_centers should be a positive integer smaller '
                         'than the number of data points')

    algorithm_b = algorithm.encode()
    if algorithm_b not in algorithm_type_ikmeans.keys():
        raise ValueError('algorithm field invalid')

    depth = max(1, math.ceil(math.log(num_leaves) / math.log(num_centers)))

    return cy_hikmeans(data, num_centers, num_leaves, depth, algorithm_b,
                       verbose)


def hikmeans_push(data, py_tree, verbose=False):
    """
    Projects the data on the HKMeans nearest elements.

    Parameters
    ----------
    data : [N, D] `uint8` `ndarray`
        Data to be projected
    py_tree : `PyHIKMNode`
        Tree structure created by ``hikmeans``
    verbose : bool, optional
        If ``True``, be verbose.

    Returns
    -------
    assignments : [N, depth] `uint32` `ndarray`
        Computed per-level assignments for ``data``
    """

    return cy_hikmeans_push(data, py_tree, verbose)



import numpy as np
from numpy.testing import assert_allclose
from cyvlfeat.quickshift.quickshift import quickshift
from cyvlfeat.quickshift.flatmap import flatmap

from cyvlfeat.test_util import lena

img = lena().astype(np.float32)
result = quickshift(img, 2, 10)
maps = result[0]


def test_flatmap_labels():
    labels, clusters = flatmap(maps)
    assert labels.shape == (512, 512)
    assert_allclose(labels[0, 0:6], [1027., 1027., 1027., 1027., 1027., 12303.],
                    rtol=1e-3)


def test_flatmap_clusters():
    labels, clusters = flatmap(maps)
    assert clusters.shape == (262144,)
    assert_allclose(clusters[0:6], [1., 1., 1., 1., 1., 56.],
                    rtol=1e-3)


def test_quickshift_medoid_maps():
    i = img.copy()
    maps, gaps, estimate = quickshift(i, kernel_size=2, max_dist=10, medoid=True)
    assert maps.shape == (512, 512)
    assert_allclose(maps[0:5, 0], [514., 1026., 1026., 1538., 1538.],
                    rtol=1e-3)


def test_quickshift_medoid_gaps():
    i = img.copy()
    maps, gaps, estimate = quickshift(i, kernel_size=2, max_dist=10, medoid=True)
    assert gaps.shape == (512, 512)
    assert_allclose(gaps[0:5, 0], [228071.506, 290406.801, 323886.572, 323339.597,
                                   293392.239], rtol=1e-3)


def test_quickshift_medoid_estimate():
    i = img.copy()
    maps, gaps, estimate = quickshift(i, kernel_size=2, max_dist=10, medoid=True)
    assert estimate.shape == (512, 512)
    assert_allclose(estimate[0:5, 0], [8.699, 11.0754, 12.350, 12.322, 11.190],
                    rtol=1e-3)


def test_quickshift_quick_maps():
    i = img.copy()
    maps, gaps, estimate = quickshift(i, kernel_size=2, max_dist=10)
    assert maps.shape == (512, 512)
    assert_allclose(maps[0:5, 0], [2., 514., 1026., 1025., 1537.],
                    rtol=1e-3)


def test_quickshift_quick_gaps():
    i = img.copy()
    maps, gaps, estimate = quickshift(i, kernel_size=2, max_dist=10)
    assert gaps.shape == (512, 512)
    assert_allclose(gaps[0:6, 3], [1., 1., 1., 1.4142, 1., 2.2360],
                    rtol=1e-3)


def test_quickshift_quick_estimate():
    i = img.copy()
    maps, gaps, estimate = quickshift(i, kernel_size=2, max_dist=10)
    assert estimate.shape == (512, 512)
    assert_allclose(estimate[0:5, 0], [8.699, 11.0754, 12.350, 12.322, 11.190],
                    rtol=1e-3)



import numpy as np
from .cykmeans import cy_ikmeans, cy_ikmeans_push, algorithm_type_ikmeans


def ikmeans(data, num_centers, algorithm="LLOYD", max_num_iterations=200,
            verbose=False):
    """
    Integer K-means

    Parameters
    ----------
    data : [N, D] `uint8` `ndarray`
        Data to be clustered
    num_centers : `int`
        Number of clusters (leaves) per level
    algorithm : {'LLOYD', 'ELKAN'}, optional
        Algorithm to be used for clustering.
    max_num_iterations : `int`, optional
        Maximum number of iterations before giving up (the algorithm
        stops as soon as there is no change in the data to cluster
        associations).
    verbose : bool, optional
        If ``True``, be verbose.

    Returns
    -------
    (centers, assignments) : ([num_centers, D] `int32` `ndarray`, [N,] `uint32` `ndarray`)
        Computed centers of the clusters and their assignments
    """

    assert isinstance(data, np.ndarray)
    assert isinstance(num_centers, int)
    assert isinstance(verbose, bool)
    if data.ndim != 2:
        raise ValueError('Data should be a 2-D matrix')

    if data.dtype != np.uint8:
        raise ValueError('Data should be uint8')

    if num_centers > data.shape[0]:
        raise ValueError('num_centers should be a positive integer smaller '
                         'than the number of data points')

    algorithm_b = algorithm.encode()
    if algorithm_b not in algorithm_type_ikmeans.keys():
        raise ValueError('algorithm field invalid')

    if (not isinstance(max_num_iterations, int)) or max_num_iterations <= 0:
        raise ValueError('max_num_iterations should be a positive integer')

    return cy_ikmeans(data, num_centers, algorithm_b, max_num_iterations,
                      verbose)


def ikmeans_push(data, centers):
    """
    Projects the data on the KMeans nearest elements (similar to
    kmeans_quantize but for integer data).

    Parameters
    ----------
    data : [N, D] `uint8` `ndarray`
        Data to be projected to the centers assignments
    centers : [K, D] `int32` `ndarray`
        Centers positions

    Returns
    -------
    assignments : [N,] `uint32` `ndarray`
        Assignments of the data points to their respective clusters indice.
    """
    assert isinstance(data, np.ndarray)

    if data.ndim != 2:
        raise ValueError('Data should be a 2-D matrix')
    if data.dtype != np.uint8:
        raise ValueError('Data should be uint8')
    if centers.ndim != 2:
        raise ValueError('Centers should be a 2-D matrix')
    if centers.dtype != np.int32:
        raise ValueError('Centers should be int32')

    return cy_ikmeans_push(data, centers)



from cyvlfeat.fisher import fisher
import numpy as np
from numpy.testing import assert_allclose


def test_fisher_dimension():
    N = 1000
    K = 512
    D = 128

    x = np.random.uniform(size=(N, D)).astype(np.float32)
    means = np.random.uniform(size=(K, D)).astype(np.float32)
    covariances = np.random.uniform(size=(K, D)).astype(np.float32)
    priors = np.random.uniform(size=(K,)).astype(np.float32)
    priors /= np.linalg.norm(priors)
    enc = fisher(x, means, covariances, priors, verbose=True)

    expected = 2 * K * D
    observed = len(enc)
    assert(expected == observed)


def _compute_posteriors(diff, covars, prior):
    q = (diff ** 2).sum(-1)  # Mahalanobis distances to cluster centers
    # Compute the log probability of samples knowing each cluster
    q = -0.5 * (q + np.log(covars).sum(-1) +
                diff.shape[1] * np.log(2 * np.pi))
    # Compute the joint probability of xi and j-th cluster
    # p(xi, cj) = p(xi|cj) * p(cj)
    q += np.log(prior)
    # Normalize values to avoid numerical problems
    qmax = q.max(-1)
    q -= qmax[:, None]
    # Compute the posteriors
    q = np.exp(q)
    # normalize
    q /= q.sum(-1)[:, np.newaxis]
    return q


def _fisher_encoding(X, means, covars, priors):
    diff = (X[:, None] - means[None]) / np.sqrt(covars)
    q = _compute_posteriors(diff, covars, priors)
    u = (diff * q[..., np.newaxis]).mean(0) / np.sqrt(priors)[:, np.newaxis]
    v = ((diff ** 2 - 1) * q[..., np.newaxis]).mean(0) / \
        np.sqrt(2 * priors)[:, np.newaxis]
    return np.hstack([u.ravel(), v.ravel()])


def test_fisher_encoding():
    np.random.seed(1)
    X = np.random.randn(1000, 2)
    X[500:] *= (2, 3)
    X[500:] += (4, 4)

    mu = np.asarray(
        [[0., 0.], [4., 4.]]
    )

    sigma2 = np.asarray(
        [[1.0, 1.0], [2.0, 3.0]]
    )

    prior = np.asarray(
        [0.6, 0.4]
    )

    expected_enc = _fisher_encoding(X, mu, sigma2, prior)
    observed_enc = fisher(X, mu, sigma2, prior, verbose=True)

    assert_allclose(expected_enc, observed_enc, rtol=1e-4)

    observed_enc = fisher(X.astype("float32"),
                          mu.astype("float32"),
                          sigma2.astype("float32"),
                          prior.astype("float32"), verbose=True)

    assert_allclose(expected_enc, observed_enc, rtol=1e-4)
    assert observed_enc.dtype == "float32"



import cyvlfeat.fisher
import cyvlfeat.generic
import cyvlfeat.gmm
import cyvlfeat.hog
import cyvlfeat.kmeans
import cyvlfeat.quickshift
import cyvlfeat.sift


from ._version import __version__
del _version



from __future__ import division
from cyvlfeat.hog import hog
import numpy as np
from numpy.testing import assert_allclose
from cyvlfeat.test_util import lena

        
img = lena().astype(np.float32) / 255.
half_img = img[:, :256]


def test_hog_cell_size_32_uoctti():
    i = img.copy()
    output = hog(i, 32)
    assert output.dtype == np.float32
    assert output.shape == (16, 16, 31)
    assert_allclose(output[:2, :2, 0],
                    np.array([[0.104189, 0.056746],
                              [0.051333, 0.03721]]), rtol=1e-4)


def test_hog_cell_size_32_uoctti_4_orientations():
    i = img.copy()
    output = hog(i, 32, n_orientations=4)
    assert output.dtype == np.float32
    assert output.shape == (16, 16, 16)
    assert_allclose(output[:2, :2, 0],
                    np.array([[0.158388,  0.085595],
                              [0.078716,  0.062816]]), rtol=1e-5)


def test_hog_cell_size_32_uoctti_non_square():
    i = half_img.copy()
    output = hog(i, 32)
    assert output.dtype == np.float32
    assert output.shape == (16, 8, 31)
    assert_allclose(output[:2, :2, 0],
                    np.array([[0.163912, 0.167787],
                              [0.086294, 0.079365]]), rtol=1e-5)


def test_hog_cell_size_32_dalaltriggs():
    i = img.copy()
    output = hog(i, 32, variant='DalalTriggs')
    assert output.dtype == np.float32
    assert output.shape == (16, 16, 36)
    assert_allclose(output[:2, :2, 0],
                    np.array([[0.139408, 0.093407],
                              [0.070996, 0.065033]]), rtol=1e-5)


def test_hog_cell_size_32_dalaltriggs_4_orientations():
    i = img.copy()
    output = hog(i, 32, n_orientations=4, variant='DalalTriggs')
    assert output.dtype == np.float32
    assert output.shape == (16, 16, 16)
    assert_allclose(output[:2, :2, 0],
                    np.array([[0.2,      0.154738],
                              [0.109898, 0.108115]]), rtol=1e-5)


def test_hog_cell_size_32_dalaltriggs_non_square():
    i = half_img.copy()
    output = hog(i, 32, variant='DalalTriggs')
    assert output.dtype == np.float32
    assert output.shape == (16, 8, 36)
    assert_allclose(output[:2, :2, 0],
                    np.array([[0.2,       0.2],
                              [0.144946,  0.192144]]), rtol=1e-5)


def test_hog_cell_size_32_dalaltriggs_bilinear_interpolation():
    i = img.copy()
    output = hog(i, 32, variant='DalalTriggs', bilinear_interpolation=True)
    assert output.dtype == np.float32
    assert output.shape == (16, 16, 36)
    assert_allclose(output[:2, -2:, 0],
                    np.array([[0.01523,  0.017774],
                              [0.012941, 0.012733]]), rtol=1e-4)



from __future__ import division
from cyvlfeat.vlad.vlad import vlad
import numpy as np
from numpy.testing import assert_allclose


def test_vlad_dimension():
    N = 1000
    K = 512
    D = 128

    x = np.random.uniform(size=(D, N)).astype(np.float32)
    means = np.random.uniform(size=(D, K)).astype(np.float32)
    assignments = np.random.uniform(size=(K, N)).astype(np.float32)
    enc = vlad(x, means, assignments)

    expected = K * D
    observed = len(enc)
    assert (expected == observed)


def test_vlad_encoding():
    N = 21
    K = 3
    D = 5
    x = np.zeros(D * N, dtype=np.float32)

    # X has one column per data vector
    for i in range(D * N):
        x[i] = i
    x = x.reshape(D, N)

    # mean has same no. of rows as X and one column per cluster
    mu = np.zeros(D * K, dtype=np.float32)
    for i in range(D * K):
        mu[i] = i
    mu = mu.reshape(D, K)

    # same no. of rows as no. of clusters and as many columns as X
    assign = np.zeros(K * N, dtype=np.float32)
    for i in range(K * N):
        assign[i] = i
    assign = assign.reshape(K, N)
    fraction = K * N
    assign *= (1 / fraction)

    observed_enc = vlad(x, mu, assign)

    # expected result obtained from running vl_vlad_encode from a C program
    expected_enc = np.array([0.27231345, 0.27231345, 0.27231348, 0.27231345, 0.27231342, 0.25836566,
                             0.25836569, 0.25836569, 0.25836569, 0.25836572, 0.24308957, 0.2430896,
                             0.24308957, 0.24308956, 0.24308957], dtype=np.float32)
    assert_allclose(expected_enc, observed_enc)



import numpy as np


def flatmap(maps):
    """
    Flatten a tree, assigning the label of the root to each node.

    Parameters
    ----------
    maps : [H, W] `float64` `ndarray`.
        Array of the same size of image.
        `maps` as returned by `quickshift`.
        `flatmap` labels each tree of the forest contained in `map`.


    Returns
    -------
    labels : contains the linear index of the root node in `map`.

    clusters : contains a label between 1 and the number of clusters.
    
    Example
    -------
    >>> import numpy as np
    >>> from cyvlfeat.quickshift.quickshift import quickshift
    >>> from cyvlfeat.quickshift.flatmap import flatmap
    >>> from cyvlfeat.test_util import lena
    >>> img = lena().astype(np.float32)
    >>> maps, gaps, estimate = quickshift(img,kernel_size=2,max_dist=10)
    >>> labels, clusters = flatmap(maps)
    """
    maps_shape = maps.shape[0] * maps.shape[1]
    root = np.ones(maps_shape, dtype=np.float64)
    maps_re = maps.reshape(maps_shape)

    # follow the parents list to the root nodes (where nothing changes)
    while 1:
        for i in range(maps_shape):
            val = int(maps_re[i] - 1)
            root[i] = maps_re[val]
        if np.array_equal(maps_re, root):
            break
        # Copy back to maps_re.
        maps_re[:] = root[:]

    labels = maps_re.reshape((maps.shape[0], maps.shape[1]))
    C = np.unique(labels, return_inverse=True)
    clusters = np.array(C[1], dtype=np.float64)
    clusters += 1

    return labels, clusters

